#include "chrono.h"
#include <memory.h>
#include <math.h>
#include "vector.h"

const double PI2 = 6.283185307179586476925286766559;

/*
The order of the symbols is:
solid rock, vertical wall, horizontal wall, upper left corner, upper right corner, lower left corner, lower right corner,
cross wall, upward T wall, downward T wall, leftward T wall, rightward T wall, vertical open door,
horizontal open door, vertical closed door, horizontal closed door, tree, floor of a room, corridor,
stairs up, stairs down, shallow water, deep water, ice, lava */

  
CharData CMap::MapChars[] = 
{ {' ',GREY}, {179,GREY}, {196,GREY}, {218,GREY}, {191,GREY}, {192,GREY}, {217,GREY},
  {197,GREY}, {193,GREY}, {194,GREY}, {180,GREY}, {195,GREY}, {254,BROWN},
  {254,BROWN}, {'+',BROWN}, {'+',BROWN}, {231,GREEN},/*241*/  {250,GREY}, {176,GREY},
  {'<',GREY},  {'>',GREY},  {249,BLUE}, {247,BLUE}, {249,LTCYAN}, {247,RED}
};

CharData CMap::LitMapChars[] = 
{ {' ',GREY}, {179,GREY}, {196,GREY}, {218,GREY}, {191,GREY}, {192,GREY}, {217,GREY},
  {197,GREY}, {193,GREY}, {194,GREY}, {180,GREY}, {195,GREY}, {254,BROWN},
  {254,BROWN}, {'+',BROWN}, {'+',BROWN}, {231,GREEN},/*241*/ {249,WHITE}, {176,WHITE},
  {'<',GREY},  {'>',GREY},  {249,BLUE}, {247,BLUE}, {249,LTCYAN}, {247,RED}
};

byte CMap::MapFlags[] =
{ NONE,   ISWALL, ISWALL, ISWALL, ISWALL, ISWALL, ISWALL,
  ISWALL, ISWALL, ISWALL, ISWALL, ISWALL, ISPASSABLE|ISDOOR,
  ISPASSABLE|ISDOOR, ISDOOR, ISDOOR,  NONE, ISPASSABLE, ISPASSABLE,
  ISPASSABLE, ISPASSABLE, ISPASSABLE, NONE, ISPASSABLE, NONE
};

CMap::CMap()
{ Rooms  = NULL;
  Stairs = NULL;
  numRooms = maxRooms = numStairs = maxStairs = 0;
}

CMap::~CMap()
{ delete[] Rooms;
  delete[] Stairs;
}

int CMap::init()
{ memset(Layout, 0, sizeof(Layout));
  memset(DisVis, 0, sizeof(DisVis));
  memset(Lit,    0, sizeof(Lit));
  memset(Outbuf, 0, sizeof(Outbuf));
  numRooms = numStairs = 0;
  return 0;
}

void CMap::deinit()
{
}

int CMap::load(FILE *file)
{ return 0;
}

int CMap::save(FILE *file)
{ return 0;
}

Coord CMap::freeSpace(bool roomonly)
{ int i;
  if(roomonly)
  { while(true)
    { Room rm = Rooms[Rnd.Rand(numRooms-1)];
      for(i=0;i<20;i++)
      { int x = Rnd.Rand(rm.area.left+1, rm.area.right-1),
            y = Rnd.Rand(rm.area.top+1, rm.area.bottom-1);
        if(layAt(x, y) == MT_ROOMFLOOR) return Coord(x, y);
      }
    }
  }
  else
  { while(true)
    { i = Rnd.Rand(MAPWID*MAPHEI-1);
      if(isPassable(Layout[i])) return Coord(i%MAPWID, i/MAPWID);
    }
  }
}

void CMap::render()
{ for(int j=0,i=0;i<MAPWID*MAPHEI;j+=2,i++)
  { byte b = DisVis[i];
    if(!b) continue;
    if(b == 1)
    { Outbuf[j]   = MapChars[Layout[i]].b.c;
      Outbuf[j+1] = MapChars[Layout[i]].b.attr;
    }
    else
    { Outbuf[j]   = LitMapChars[Layout[i]].b.c;
      Outbuf[j+1] = LitMapChars[Layout[i]].b.attr;
    }
  }
}

void CMap::draw()
{ Con.putBlock(0, TOPAREA, MAPWID, MAPHEI, Outbuf);
}

int CMap::layAdj(int x, int y, byte b)
{ byte *lay;
  int num=0;
  int ox, xe, ye, nx=x, ny=y, inc;

  xe = (x==MAPWID-1 ? x : x+1), ye = (y==MAPHEI-1 ? y : y+1);
  if(x) x--;
  inc = MAPWID-(xe-(ox=x)+1);
  if(y) y--;

  lay = Layout+y*MAPWID+x;
  for(; y<=ye; y++)
  { for(x=ox; x<=xe; x++) if(*lay++ == b && (x!=nx || y!=ny)) num++;
    lay += inc;
  }
  return num;
}

int CMap::doorAdj(int x, int y)
{ byte *lay;
  int num=0;
  int ox, xe, ye, nx=x, ny=y, inc;

  xe = (x==MAPWID-1 ? x : x+1), ye = (y==MAPHEI-1 ? y : y+1);
  if(x) x--;
  inc = MAPWID-(xe-(ox=x)+1);
  if(y) y--;

  lay = Layout+y*MAPWID+x;
  for(; y<=ye; y++)
  { for(x=ox; x<=xe; x++) if(isDoor(*lay++) && (x!=nx || y!=ny)) num++;
    lay += inc;
  }
  return num;
}

int CMap::addRoom()
{ Room r;
  r.con = false;

  for(int t=0;t<MGTRIES;)
  { r.area.left   = Rnd.Rand(0, 76);
    r.area.top    = Rnd.Rand(0, 76);
    r.area.right  = Rnd.Rand(r.area.left+3, 79);
    r.area.bottom = Rnd.Rand(r.area.top +3, 79);
    if(r.width() > MGMAXWID || r.height() > MGMAXHEI) continue;
    r.area.left   = r.area.left   * MAPWID / 80;
    r.area.right  = r.area.right  * MAPWID / 80;
    r.area.top    = r.area.top    * MAPHEI / 80;
    r.area.bottom = r.area.bottom * MAPHEI / 80;
    if(r.height() < 5 || r.width() < 5) continue;
    for(int i=0; i<numRooms; i++)
      if(r.intersects(Rooms+i, 2)) break;
    if(i==numRooms) break;
    t++;
  }
  if(t==MGTRIES) return 1;

  if(numRooms == maxRooms)
  { Room *nr = new Room[maxRooms+=10];
    memcpy(nr, Rooms, numRooms*sizeof(Room));
    Rooms = nr;
  }
  Rooms[numRooms++] = r;
  return 0;
}

int CMap::addStairs(bool up)
{ Coord pos = freeSpace(true);
  if(up) layTo(pos.x, pos.y, MT_STAIRUP);
  else layTo(pos.x, pos.y, MT_STAIRDOWN);
  if(numStairs == maxStairs)
  { Stair *s = new Stair[maxStairs+=5];
    if(numStairs) memcpy(s, Stairs, numStairs*sizeof(Stair));
    Stairs = s;
  }
  Stairs[numStairs].pos = pos;
  Stairs[numStairs].up  = up;
  numStairs++;
  return 0;
}

void CMap::updateVis(Creature *sees)
{ Vector<float> vect(0, -1);
  float ang=0.0, ai, sight=(float)sees->sightdist, light=(float)sees->lightrad,
        sx=(float)(sees->pos.x + 0.5), sy=(float)(sees->pos.y + 0.5);

#ifdef SAFERAYS
  ai=256/(float)(sees->sightdist*8);  
#else
  if(sees->sightdist <= 5) ai=256/(float)(sees->sightdist*6.5);
  else if(sees->sightdist <= 10) ai=256/(float)(sees->sightdist*7);
  else ai=256/(float)(sees->sightdist*8);
#endif

  for(int i=0;i<MAPWID*MAPHEI;i++) if(DisVis[i]) DisVis[i] = 1;

  for(;ang<256.0;ang+=ai)
  { Vector<float> rot = RotatedZ(vect, (int)(ang+0.5));
    Vector<float> rot2 = rot * light;
    rot *= sight;
    traceLine(sx, sy, sx+rot.X, sy+rot.Y, sx+rot2.X, sy+rot2.Y);
  }
}

void CMap::traceLine(float sx, float sy, float dx, float dy, float dx2, float dy2)
{ float xd = dx-sx, yd=dy-sy, xd2=dx2-sx, yd2=dy2-sy, ax, ay;
  ax = (float)fabs(xd), ay = (float)fabs(yd);

  if(ax > ay)
  { float xi = (float)(xd < 0 ? -0.5 : 0.5), yi = yd/(ax*2);
    int num = (int)(ax*2+0.5), light=(int)(fabs(xd2)*2+0.5);
    if((num -= light) < 0) num = 0;
    while(light--)
    { sx += xi, sy += yi;
      int i = (int)sy*MAPWID+(int)sx;
      DisVis[i] = 2;
      if(!isPassable(Layout[i])) return;
    }
    while(num--)
    { sx += xi, sy += yi;
      int i = (int)sy*MAPWID+(int)sx;
      if(Lit[i]) DisVis[i] = 2;
      else return;
      if(!isPassable(Layout[i])) return;
    }
  }
  else
  { float xi = xd/(ay*2), yi = (float)(yd < 0 ? -0.5 : 0.5);
    int num = (int)(ay*2+0.5), light=(int)(fabs(yd2)*2+0.5);
    if((num -= light) < 0) num = 0;
    while(light--)
    { sx += xi, sy += yi;
      int i = (int)sy*MAPWID+(int)sx;
      DisVis[i] = 2;
      if(!isPassable(Layout[i])) return;
    }
    while(num--)
    { sx += xi, sy += yi;
      int i = (int)sy*MAPWID+(int)sx;
      if(Lit[i]) DisVis[i] = 2;
      else return;
      if(!isPassable(Layout[i])) return;
    }
  }
}
